import { Node, PropertyHint, Variant } from 'godot'
import { export_, onready as g_onready } from 'godot.annotations'

const get_node_from_key = (key: string, node: Node) => {
  const path = node.get(key)

  if (path == null) {
    console.warn(`${node.get_name()}::${key} is null, can't export node`)
    return
  }

  return node.get_node_or_null(path)
}

export function export_scene() {
  return export_resource('PackedScene')
}

export function export_resource(type: string = '') {
  return export_(Variant.Type.TYPE_STRING, { hint: PropertyHint.PROPERTY_HINT_RESOURCE_TYPE, hint_string: type })
}

export function export_expression() {
  return export_(Variant.Type.TYPE_STRING, { hint: PropertyHint.PROPERTY_HINT_EXPRESSION })
}

export function export_node<T extends Node>(defaultNode?: string) {
  return function(target: T, key: string) {
    const meta_key = `_${key}`
    Reflect.set(target, meta_key, defaultNode || '')

    export_(Variant.Type.TYPE_NODE_PATH)(target, meta_key)
    onready((node: Node) => get_node_from_key(meta_key, node))(target, key)
  }
}

export function onready<T extends Node, ValueType>(
  evaluator: ((node: T) => ValueType) | string
) {
  if (typeof evaluator === 'string') {
    return g_onready(evaluator)
  }

  return function(target: T, property: string) {
    const handlersKey = '$onready:handlers'

    let handlers = Reflect.get(target, handlersKey) as {
      property: string
      evaluator: (node: T) => ValueType
    }[]

    if (!handlers) {
      handlers = []
      Reflect.set(target, handlersKey, handlers)

      const old_ready: () => void = Reflect.get(target, '_ready') || (() => { })

      Reflect.defineProperty(target, '_ready', {
        value: function() {
          for (const { evaluator, property } of handlers) {
            this[property] = evaluator?.(this)
          }

          old_ready.call(this)
        }
      })
    }

    handlers.push({ evaluator, property })
  }
}


