import { GArray, GError, Node, Resource, ResourceLoader, ResourceSaver } from 'godot'
import { GodotError } from './godot_error'

export class ResourceLoadError extends Error { }

export default class AsyncResourceHandler extends Node {
  private static _instance: AsyncResourceHandler
  static readonly SaveFlags = ResourceSaver.SaverFlags
  static readonly CacheMode = ResourceLoader.CacheMode

  static get instance() {
    return this._instance
  }

  _ready(): void {
    AsyncResourceHandler._instance = this
  }

  static save<T extends Resource>(path: string, data: T, flags: ResourceSaver.SaverFlags): Promise<void> {
    return this._instance.save(path, data, flags)
  }

  save<T extends Resource>(path: string, data: T, flags: ResourceSaver.SaverFlags): Promise<void> {
    return new Promise((resolve, reject) => {
      const err = ResourceSaver.save(data, path, flags)

      if (err === GError.OK) {
        resolve()
      } else {
        reject(err)
      }
    })
  }

  static load<T extends Resource>(path: string, hint_string: string = '', use_sub_threads?: boolean, cache_mode?: ResourceLoader.CacheMode): Promise<T> {
    return this._instance.load(path, hint_string, use_sub_threads, cache_mode)
  }

  load<T extends Resource>(path: string, hint_string: string = '', use_sub_threads: boolean = false, cache_mode: ResourceLoader.CacheMode = ResourceLoader.CacheMode.CACHE_MODE_REUSE): Promise<T> {
    return new Promise(async (resolve, reject) => {
      const err = ResourceLoader.load_threaded_request(path, hint_string, use_sub_threads, cache_mode)

      if (err != GError.OK) {
        return reject(new GodotError(err))
      }

      const get_status = () => ResourceLoader.load_threaded_get_status(path)

      let status = get_status()
      while (status === ResourceLoader.ThreadLoadStatus.THREAD_LOAD_IN_PROGRESS) {
        await this.get_tree().process_frame.as_promise()
        status = get_status()
      }
      if (status === ResourceLoader.ThreadLoadStatus.THREAD_LOAD_LOADED) {
        return resolve(ResourceLoader.load_threaded_get(path) as T)
      } else {
        return reject(new ResourceLoadError())
      }
    })
  }

  static get_progress(path: string): number {
    return this._instance.get_progress(path)
  }

  get_progress(path: string): number {
    const arr = new GArray()
    ResourceLoader.load_threaded_get_status(path, arr)
    return arr.get_indexed(0)
  }
}

