import { GDictionary, Node, PackedScene, Resource, ResourceLoader, ResourceSaver } from 'godot'
import AsyncResourceHandler from './async_resource_handler'
import { has_method } from './node'

interface Serializable {
  on_save_game(save_data: GameState): void
  on_before_load_game?(): void
  on_load_game<T extends SaveState>(state: T): void
}

interface SaveState extends Resource {
  scene: string
}

class PlayerState extends Resource implements SaveState {
  condition: undefined = undefined
  inventory: undefined = undefined
  scene: string = ''
}

class ProgressState extends Resource {
  variables?: GDictionary
}

export class GameState extends Resource {
  player?: PlayerState
  progress?: ProgressState
  world?: SaveState[]
}

const StateMethod = {
  Save: 'on_save_game',
  BeforeLoad: 'on_before_load_game',
  Load: 'on_load_game'
} as const

type StateMethod = typeof StateMethod[keyof typeof StateMethod]

export default class SaveSerializer extends Node {
  private static _save_group: string = 'save'

  private static _instance: SaveSerializer

  _ready() {
    SaveSerializer._instance = this
  }

  static save() {
    this._instance.save()
  }

  private _announce(method: StateMethod, ...args: any) {
    this.get_tree().call_group(
      SaveSerializer._save_group,
      method,
      ...args
    )
  }

  save() {
    const game_state: GameState = new GameState()

    this._announce(StateMethod.Save, game_state)

    ResourceSaver.save(game_state, 'user://save1.tres')
  }

  private async _restore_node<T extends SaveState>(state: T, parent: Node) {
    const scene: PackedScene = await AsyncResourceHandler.load(state.scene, 'PackedScene')
    const node = scene.instantiate()
    parent.add_child(node)

    if (has_method(StateMethod.Load, node)) {
      // TODO: type check for state method
    }

  }

  async load() {
    const loaded_state: GameState = await AsyncResourceHandler.load('user://save1.tres')
    this._announce(StateMethod.BeforeLoad)
    // TODO: pass world root as 2nd param
    const node_promises = loaded_state.world?.map(state => this._restore_node(state, this)) ?? []
    return Promise.all(node_promises)
  }
}

